@@ -103,13 +103,23 @@
     (box (&AST->datum (unbox stx))))
    (else stx)))
 
+;; TODO: change to version check when gambit releases 4.9.4
+(eval-if-bound ##convert-case
+  (define (get-readenv port)
+    (##make-readenv port
+                    (current-readtable)
+                    _gx#wrap-syntax
+                    _gx#unwrap-syntax
+                    #f '() #f))
+  (define (get-readenv port)
+    (##make-readenv port
+                    (current-readtable)
+                    _gx#wrap-syntax
+                    _gx#unwrap-syntax
+                    #f #f)))
+
 (define (read-syntax #!optional (in (current-input-port)))
-  (let ((e (##read-datum-or-eof
-            (##make-readenv in
-                            (current-readtable)
-                            _gx#wrap-syntax
-                            _gx#unwrap-syntax
-                            #f #f))))
+  (let ((e (##read-datum-or-eof (get-readenv in))))
     (if (eof-object? (&AST-e e))
       (&AST-e e)
       e)))
